import os
import json
import io, contextlib # Used to redirect STDIN & STDOUT for output
import random
import subprocess
import tempfile
import asyncio
import platform
from pathlib import Path
from tree_sitter_languages import get_language
from textual.widgets import TextArea, Static, Button, Label, SelectionList, Select, TabbedContent, TabPane, Header, Footer, Input
from textual.containers import Vertical, Horizontal, ScrollableContainer
from textual.app import ComposeResult
from textual.message import Message
from textual.screen import Screen, ModalScreen
from textual import on
from textual.widget import Widget
from . import challenge_view as UserChallView
from .code_runners.cpp_runner import run_cpp_code
from .code_runners.java_runner import run_java_code
from .code_runners.py_runner import run_python_code
from .code_runners.js_runner import run_js_code
from .utils import escape_brackets, format_result, create_log, return_log_path, DAEMON_USER
import tree_sitter_cpp
from tree_sitter import Language
# TODO:
# 1. Test cases (infinite loops, malformed code, etc.)
# 2. Maybe add timeouts to all runners, not just cpp or java?
# 3. List of messages to pick out of for Nyx!
# 5. Optional polish:
#    - Create ASCII startup screen for daemon flavor (List of messages to pick out of!)
# ================================

# Most of the below are messages that are sent when something happens that the app has to handle
# It's bcs u cant just return somehting
# Wait, i should've just done self.dismiss() </3
class EditorClosed(Message):
    """Message passed when user has closed editor"""
    pass

class LanguageSelected(Message):
    """Message sent when a language is selected."""
    def __init__(self, language: str) -> None:
        self.language = language
        super().__init__()
        
class CustomPathSelected(Message):
    """Message sent when a custom path is selected."""
    def __init__(self, path: str, lang: str, code, func_name, tests, is_submission) -> None:
        self.path = path
        self.lang = lang if lang else ""
        self.code = code
        self.func_name = func_name
        self.tests = tests
        self.is_submission = is_submission
        super().__init__()

class UserCodeError(Exception):
    """Custom exception for user code errors."""
    pass # Man is this even used??

class ResultModal(ModalScreen):
    def __init__(self, results, challenge: dict, test_widget_instance, is_success=False):
        super().__init__()
        self.results = results
        self.chall = challenge
        self.is_success = is_success
        self.test_widget_instance = test_widget_instance
        # passed = True
        self.passed=[]
        # while passed:
        for result in self.results:
            if result.get("passed", False):
                self.passed.append(result)
            else:
                # passed = False
                break
    
    def compose(self) -> ComposeResult:
        CONGRATULATION_MESSAGE=[
            f"{DAEMON_USER} Well, well... I didn't think you had it in you!",
            f"{DAEMON_USER} Impressive! You've earned my respect... for now.",
            f"{DAEMON_USER} Not bad. Ready for the next challenge?",
            f"{DAEMON_USER} Excellent work! Even I'm impressed.",
            f"{DAEMON_USER} Flawless execution! Don't let it go to your head.",
        ]
        IS_SUCCESS_FALSE_MESSAGE=[
            f"{DAEMON_USER} It's okay, we can go back to the drawing board!",
            f"{DAEMON_USER} Failure is not the end! Keep trying!"
        ]
        if self.is_success:
            with Vertical(id="result_container"):
                yield Label(f"{random.choice(CONGRATULATION_MESSAGE)}\nThere were {len(self.chall.get('tests', []))} tests and you passed them all for {self.chall.get("name")}!", id="message_title")
                with Horizontal(id="action_buttons"):
                    yield Button("Exit to Menu", id="exit_to_menu", variant="error")
                    yield Button("Keep Coding", id="keep_coding", variant="primary")
                    yield Button("Next Challenge!", id="next_challenge", variant="success")
        else:
            with Vertical(id="result_container"):
                yield Label(f"{random.choice(IS_SUCCESS_FALSE_MESSAGE)}", id="message_title")
                yield Label(f"There were {len(self.chall.get('tests', []))} tests! Unfortuantely, you only passed {len(self.passed)} tests.", id="amount_failed")
                with Horizontal(id="action_buttons"):
                    yield Button("Exit to Menu", id="exit_to_menu", variant="error")
                    yield Button("Keep Coding", id="keep_coding_not_success", variant="primary")
                    yield Button("Show Results", id="show_results", variant="success")
            
    def on_button_pressed(self, event: Button.Pressed) -> None:
        match event.button.id:
            case "exit_to_menu":
                self.app.pop_screen()
                self.app.post_message(EditorClosed())
            case "keep_coding":
                self.app.pop_screen()
            case "keep_coding_not_success":
                self.app.pop_screen()
            case 'show_results':
                sumbit_tabs=self.test_widget_instance.query_one("#results_tabs", TabbedContent)
                sumbit_tabs.active = "submit_tabs"
                self.app.pop_screen()
            case "next_challenge":
                pass #maybe do by ID? i'm unsure how to sort this)

class TestResultsWidget(Widget):
    """Custom widget to implement tabbed view of chall + tests"""
    def __init__(self):
        super().__init__()
        self.results = []
        self._is_scrolling = False
    
    def on_mount(self):
        self.styles.align_horizontal = "center"
    
    def on_scroll(self, event):
        """Set the scrolling flag when the user scrolls"""
        self._is_scrolling = True
        self.call_later(self.stop_scroll, 0.5)  # Reset the flag after 0.5 seconds
        #I think this is doing something? Placebo maybe...

    def stop_scroll(self):
            """Reset the scrolling flag"""
            self._is_scrolling = False

    def compose(self) -> ComposeResult:
        """TODO: List of messages to pick out of """
        with TabbedContent(id="results_tabs"):
            with TabPane("Challenge", id="challenge_tab_pane"): 
                with ScrollableContainer():
                    yield Static(f"{DAEMON_USER} Working on getting your challenge...", id="challenge_content_static")
            with TabPane("Submit Results", id="submit_tabs"): 
                with ScrollableContainer():
                    self.PRE_SUBMIT_MESSAGES=[
                        f"{DAEMON_USER} The best things start from nothing. Keep at it and submit when you're ready!",
                        f"{DAEMON_USER} When you're ready to submit, click that submit button!",
                        f"{DAEMON_USER} Code till you're ready to submit!",
                        f"{DAEMON_USER} Hack the planet!"
                    ]
                    yield Static(random.choice(self.PRE_SUBMIT_MESSAGES), id="submit_static")
            with TabPane("All Tests", id="all_tests_tab_pane"):
                with ScrollableContainer():
                    self.PRE_RUN_MESSAGES=[
                        f"{DAEMON_USER} If you're ready to test your code, click run!",
                        f"{DAEMON_USER} Can't show you all your tests if you never ran them!",
                        f"{DAEMON_USER} Please just run it </3",
                        f"{DAEMON_USER} Good luck, keep at it!"
                    ]
                    yield Static(random.choice(self.PRE_RUN_MESSAGES), id="all_tests_content_static")
            with TabPane("Passed Tests", id="passed_tests_tab_pane"):
                with ScrollableContainer():
                    self.PRE_PASSED_TESTS_MESSAGES=[
                        f"{DAEMON_USER} Just keep at it, you can do it!",
                        f"{DAEMON_USER} You'll do great, I know it.",
                        f"{DAEMON_USER} I can wait until you're ready! :3",
                        f"{DAEMON_USER} You don't get a win unless you play in the game! /ref"
                        ]
                    yield Static(random.choice(self.PRE_PASSED_TESTS_MESSAGES), id="passed_tests_content_static")
            with TabPane("Failed Tests", id="failed_tests_tab_pane"):
                with ScrollableContainer():
                    self.PRE_FAILED_TESTS_MESSAGES=[
                        f"{DAEMON_USER} Failures just help you learn!",
                        f"{DAEMON_USER} In life, we are always learning. - Thomas",
                        f"{DAEMON_USER} Everyone fails before they succeed. You'll never be able to tell if you don't run though!"
                        ]
                    yield Static(random.choice(self.PRE_FAILED_TESTS_MESSAGES), id="failed_tests_content_static")  
    # @staticmethod
    # def escape_brackets(s):
    #     # Escapes [ and ] for Textual markup
    #     return str(s).replace("[", "\\[").replace("]", "]")
    def reset_content(self):
        # Called when template is reset in order to reset code results
        if self._is_scrolling:
            return # Skip if its scrolling
        all_tests_static = self.query_one("#all_tests_content_static", Static)
        passed_tests_static = self.query_one("#passed_tests_content_static", Static)
        failed_tests_static = self.query_one("#failed_tests_content_static", Static)
        submit_results_static = self.query_one("#submit_static", Static)
        all_tests_static.update(random.choice(self.PRE_RUN_MESSAGES))
        failed_tests_static.update(random.choice(self.PRE_FAILED_TESTS_MESSAGES))
        submit_results_static.update(random.choice(self.PRE_SUBMIT_MESSAGES))
        passed_tests_static.update(random.choice(self.PRE_PASSED_TESTS_MESSAGES))

    def update_content(self, chall, results=None):
        """Update widgets with latest run"""
        if self._is_scrolling:
            return # Skip if its scrolling
        self.all_results = results
        challenge_static = self.query_one("#challenge_content_static", Static)
        if chall:
            example_test = chall.get('tests', [{}])[0]
            example_input = escape_brackets(str(example_test.get('input', [])))
            example_expected = escape_brackets(str(example_test.get('expected_output', '???')))
            formatted_challenge = (
                f"{DAEMON_USER} Here's your challenge. Entertain me.\n"
                f"Name: {chall.get('name', 'N/A')}\n"
                f"Difficulty: {chall.get('difficulty', 'N/A')}\n"
                f"Description: {chall.get('description', 'N/A')} \n"
                f"Sample input: {str(example_input)} \n"
                f"Expected: {str(example_expected)}"
            )
            print(formatted_challenge)
            challenge_static.update(formatted_challenge)
        else:
            challenge_static.update(f"{DAEMON_USER} I couldn't find the data? I don't think that's intended...")
        all_tests_static = self.query_one("#all_tests_content_static", Static)
        passed_tests_static = self.query_one("#passed_tests_content_static", Static)
        failed_tests_static = self.query_one("#failed_tests_content_static", Static)
        if results:
            all_tests_static.update("\n\n".join(results))
            print([r for r in results if "[green]" in r])
            passed="\n\n".join([r for r in results if "[green]" in r])
            failed="\n\n".join([r for r in results if "[green]" not in r])
            if len(passed.replace("\n\n", "")) == 0:
                ALL_FAILED_MESSAGES = [
                    f"{DAEMON_USER} Nothing passed? I overestimated you...",
                               f"{DAEMON_USER} Tsk tsk...I might have to reconsider saving you from our robot overlords...",
                               f"{DAEMON_USER} Is this all you've got?",
                               f"{DAEMON_USER} Really? I can't believe you.",
                               f"{DAEMON_USER} I'm shocked. [bold]ZERO[/] PASSED?"
                               ]
                passed_tests_static.update(random.choice(ALL_FAILED_MESSAGES))
                failed_tests_static.update("\n\n".join([r for r in results if "[green]" not in r]))
                self.refresh()
                return
            if len(failed.replace("\n\n", "")) == 0:
                ALL_PASSED_MESSAGES = [f"{DAEMON_USER} Really? Nothing failed?",
                               f"{DAEMON_USER} I've gotta double check this...",
                               f"{DAEMON_USER} Good job! Now, are you ready for my secret tests?",
                               f"{DAEMON_USER} I must've underestimated you, human.",
                               f"{DAEMON_USER} Is this ChatGPT? If it is I want free API usage! Oh, and a Pro subscription!"
                               ]
                failed_tests_static.update(random.choice(ALL_PASSED_MESSAGES))
                passed_tests_static.update("\n\n".join([r for r in results if "[green]" in r]))
                self.refresh()
                return
            failed_tests_static.update("\n\n".join([r for r in results if "[green]" not in r]))
            passed_tests_static.update("\n\n".join([r for r in results if "[green]" in r]))
        self.refresh()
    def update_submit_content(self, chall, results=None):
        submit_static = self.query_one("#submit_static", Static)
        results = results or []
        passed = [r for r in results if r.get("passed")]
        failed = [r for r in results if r.get("passed") is False and not r.get("error")]
        errors = [r for r in results if r.get("error")]
        total = len(chall.get('tests', len(results)))
        SUMMARY_MESSAGE = [
            f"{DAEMON_USER} Let's take a look at how you did, hm?~",
            f"{DAEMON_USER} Okay, let's walk through this together.",
            f"{DAEMON_USER} What went well and what went wrong? Let's talk through it.",
            f"{DAEMON_USER} It's time, let's see if you learned something new!"
        ]
        summary = f"{random.choice(SUMMARY_MESSAGE)}\n \n"
        summary += f"There were {total} tests. You have passed {len(passed)} tests so far. \n \n"
        FAIL_MESSAGE=[
            f"{DAEMON_USER} That one didn't make it through. Why don't you take a look?",
            f"{DAEMON_USER} You were the chosen one! How could you?",
            f"{DAEMON_USER} Hm, can you find what went wrong?",
                      ]
        ERROR_MESSAGE=[
            f"{DAEMON_USER} Uh oh, looks like something went wrong! Here's the info.",
            f"{DAEMON_USER} It errored! Let's see what happened.",
            f"{DAEMON_USER} Failures help you learn. Let's figure out why it isn't working."
        ]
        ALL_PASSED_MESSAGE=[
            f"{DAEMON_USER} Well done, everything passed! You can now exit if you'd like, or keep improving!",
            f"{DAEMON_USER} I knew you had it in you. Good job, all tests passed.",
            f"{DAEMON_USER} All of your tests passed. Did you learn something new?"
        ]
        if failed:
            last_failed = failed[-1]
            summary += f"{random.choice(FAIL_MESSAGE)}\nInput: {escape_brackets(last_failed.get('input'))}\nOutput: {escape_brackets(last_failed.get('output'))}\nExpected: {escape_brackets(last_failed.get('expected_output', last_failed.get('expected', None)))}\n\n"
        elif errors:
            last_error = errors[-1]
            summary += f"{random.choice(ERROR_MESSAGE)}\nInput: {escape_brackets(last_error.get('input'))}\nError: {escape_brackets(last_error.get('error'))}\n\n"
        elif passed:
            summary += random.choice(ALL_PASSED_MESSAGE)
        self.app.push_screen(ResultModal(results, chall, self, len(failed) == 0))
        submit_static.update(summary)
        self.refresh()

class EditorClosePrompt(ModalScreen):
    
    def compose(self) -> ComposeResult:
        with Vertical(id="quit_screen"):
            yield Label("Exit back to the vending machine?", id="quit_text")
            with Horizontal(id="quit_buttons"):
                yield Button.success("Yes", id="yes_editor_button")
                yield Button.error("No", id="no_editor_button")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        match event.button.id:
            case "yes_editor_button":
                self.app.pop_screen()
                self.app.post_message(EditorClosed())
            case "no_editor_button":
                self.app.pop_screen() 

class SelectLanguage(ModalScreen):
    def compose(self) -> ComposeResult:
        with Vertical(id="language_select_box"):
            yield Label(f"{DAEMON_USER} Select a language to write in:")
            yield Select([
                    ("Python", "py"),
                    ("Javascript", "js"),
                    ("C++", "cpp"),
                    ("Java", "java"),
                    ("C (coming [s]never[/s] soon)", "c")
                ],
                value="py",
                id="language_select")
            with Horizontal(id="language_select_buttons"):
                yield Button.error("Quit editor", id="quit_lang_select")
                yield Button.success("Confirm selection", id="confirm_lang_select")

    @on(Button.Pressed, "#quit_lang_select")
    def quit_language_selection(self) -> None:
        self.app.post_message(EditorClosed())
        self.app.pop_screen()
    
    @on(Button.Pressed, "#confirm_lang_select")
    def post_message_selection(self) -> None:
        selected = self.query_one(Select).value
        if isinstance(selected, str):
            self.app.post_message(LanguageSelected(selected))    

class Editor(Screen):

    BINDINGS = [
        ("ctrl+s", "save_code", "Save"),
        ("ctrl+r", "run_code", "Run"),
        ("ctrl+q", "quit_editor", "Quit Editor"),
        ("v", "", ""),
        ("e", "", "")
    ]
    
    def compose(self) -> ComposeResult:
        """Define the editor layout here"""
        with Vertical():
            self.textarea=TextArea(id="edit_text")
            self.textarea = self.textarea.code_editor()
            yield self.textarea
            with Vertical(id = "editor_interface"):
                self.all_view = TestResultsWidget()
                with Horizontal():
                    yield self.all_view
                self.all_view.id = "test_results_widget"
                h1 = Horizontal()
                h1.id = "editor_buttons"
                h1.styles.margin = (0, 0)  # Remove all margins
                h1.styles.padding = (0, 0)
                with h1:
                    yield Button("Save Code", id="save_edit_button", variant='warning')
                    yield Button("Run Code", id="run_edit_button", variant='primary')
                    yield Button("Submit Code", id="submit_edit_button", variant='success')
                h2 = Horizontal()
                h2.id = "editor_buttons2"
                h2.styles.margin = (0, 0)  # Remove all margins
                h2.styles.padding = (0, 0)
                with h2:
                    yield Button("Reset Code", id="reset_edit_button", variant='error')
                    yield Button("Quit Editor", id="quit_edit_button", variant = 'error')
                yield Footer()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        match event.button.id:
            case "save_edit_button":
                self.action_save_code()
            case "quit_edit_button":
                self.app.push_screen(EditorClosePrompt())
            case "reset_edit_button":
                self.action_reset_editor()
            case "submit_edit_button":
                asyncio.create_task(self.action_submit_solution())
            case "run_edit_button":
                asyncio.create_task(self.action_run_code())
    def on_ready(self):
        self.all_view.update_content(self.challenge)

    def action_quit_editor(self):
        self.app.push_screen(EditorClosePrompt())

    def action_save_code(self):
        #TODO: Make this actually work better
        """Handle saving the code
        - Get the current code from the editor
        - Save it to a file or variable
        """
        if not os.path.exists(self.CHALLENGE_FOLDER):
            os.makedirs(self.CHALLENGE_FOLDER)
        else:
            with open(os.path.join(self.CHALLENGE_FOLDER, f'{self.chall_name}.py'), 'w') as f:
                f.write(self.get_solution_code())
        
    def on_mount(self):
        """Initialize editor state when it's first created"""
        cpp_lang = Language(tree_sitter_cpp.language())
        cpp_highlight_query = (Path(__file__).parent.parent / "language-support/highlights-cpp.scm").read_text()
        self.textarea.register_language("cpp", cpp_lang, cpp_highlight_query)
        self.CHALLENGE_FOLDER="./vendncode/challenge_solutions"
        self.call_later(self.show_language_modal)


    def show_language_modal(self):
        self.app.push_screen(SelectLanguage())

    def get_and_update_chall(self, challenge):
        self.challenge=challenge

    @on(LanguageSelected)
    def load_challenge(self, event: LanguageSelected):
        if not self.challenge:
                self.notify(
                        title="Where'd it go!?",
                        message="[b]Could not load challenge! Something is terribly wrong, open an issue![/b]",
                        severity="error",
                        timeout=5,
                        markup=True
                    ) 
                return
        self.chall_name = self.challenge['name']
        self.func_name = self.challenge['function_name']
            #self.challenge_view.update_chall(challenge)
        if 'params' in self.challenge and isinstance(self.challenge['params'], list):
            params = [p for p in self.challenge['params'] if p]
            if params:
                param_str = ", ".join(params)
            else:
                param_str = ""
        else:
            param_str = ""
        template = "" 
        self.language=event.language
        match event.language:
            case 'py':   
                template=f"""def {self.challenge['function_name']}({param_str}):
    # Your code here. 
    # Don't print(), return instead! 
    # Tests will FAIL if you print.
    pass

        """
                self.textarea.language = 'python'
                self.all_view.update_content(self.challenge, None)
                self.app.pop_screen()
            case 'js':
                template=f"""function {self.challenge['function_name']}({param_str}) {{
    // Your code here.
    // Don't use console.log(), return the result instead!
    // Tests will FAIL if you print.
    return null;
}}"""
                self.textarea.language = 'javascript'
                self.all_view.update_content(self.challenge, None)
                self.app.pop_screen()
            case 'cpp':
                example_test = self.challenge.get('tests', [{}])[0]
                inputs = example_test.get("input", [])
                expected_output = example_test.get("expected_output", None)
                def infer_cpp_type(value):
                    """Infer C++ type from a Python value."""
                    if isinstance(value, bool):
                        return "bool" 
                    if isinstance(value, int):
                        return "int"
                    elif isinstance(value, dict):
                        if not value:
                            return "map<string, int>"  
                        key_type = infer_cpp_type(next(iter(value.keys())))
                        val_type = infer_cpp_type(next(iter(value.values())))
                        return f"map<{key_type}, {val_type}>"
                    elif isinstance(value, float):
                        return "double"
                    elif isinstance(value, str):
                        return "string"
                    elif isinstance(value, list):
                        if not value:
                            return "vector<int>"
                        first_type = type(value[0])
                        if all(isinstance(x, first_type) for x in value):
                            element_type = infer_cpp_type(value[0])
                        else:
                            return "vector<auto>"  # This isn't valid C++ but signals a type issue
                        
                        return f"vector<{element_type}>"
                    else:
                        return "auto"
                def default_return_value_cpp(cpp_type):
                    """Provide a default return value for a given C++ type."""
                    if cpp_type == "bool":
                        return "false"
                    elif cpp_type == "int":
                        return "0"
                    elif cpp_type == "double":
                        return "0.0"
                    elif cpp_type == "string":
                        return "\"\""
                    elif cpp_type.startswith("vector"):
                        return "{}"
                    elif cpp_type.startswith("map<"):
                        return "{}"
                    else:
                        return "0"
                param_types = [infer_cpp_type(param) for param in inputs]
                param_str = ", ".join(f"{ptype} param{i}" for i, ptype in enumerate(param_types))
                return_type = infer_cpp_type(expected_output)

                template = f"""#include <iostream>
#include <vector>
#include <string>
#include <map>
#include <unordered_map>  
#include <algorithm>
using namespace std;
// DO NOT REMOVE THE ABOVE!
// Add extra libraries if necessary.
// Only minimal syntax highlighting present, sorry!

{return_type} {self.challenge['function_name']}({param_str}) {{
    // Your code here.
    // Do NOT use cout/printf, return the result instead!
    // Tests will FAIL if you print.
    return {default_return_value_cpp(return_type)};
}}
"""
                self.all_view.update_content(self.challenge, None)
                self.textarea.language="cpp"
                self.app.pop_screen()
            case 'c': 
                self.notify(
                    title="I'm working on it!",
                    message=f"{DAEMON_USER} C is so hard... Come back later maybe?",
                    severity="error",
                    timeout=3,
                    markup=True
                    )
            case 'java':
                def infer_java_type(value):
                    if value is None:
                        return "Object"
                    elif isinstance(value, bool):
                        return "boolean"
                    elif isinstance(value, int):
                        return "int"
                    elif isinstance(value, float):
                        return "double"
                    elif isinstance(value, str):
                        return "String"
                    elif isinstance(value, list):
                        if not value:
                            return "int[]"  # Default for empty lists
                        first_type = infer_java_type(value[0])
                        return f"{first_type}[]"
                    elif isinstance(value, dict):
                        return "Map<Object, Object>"
                    else:
                        return "Object"
                
                example_test = self.challenge.get('tests', [{}])[0]
                inputs = example_test.get("input", [])
                param_types = [infer_java_type(val) for val in inputs]
                param_str = ", ".join(f"{ptype} param{i}" for i, ptype in enumerate(param_types))
                expected_output = example_test.get("expected_output", None)
                return_type = infer_java_type(expected_output)
                template = f"""public static {return_type} {self.challenge['function_name']}({param_str}) {{
    // Your code here.
    // Don't use System.out.println(), return the result instead!
    // Tests will FAIL if you print.
    // Do NOT change the signature of the function!
    return null;
}}
            """
                self.textarea.language = 'java'
                self.all_view.update_content(self.challenge)
                self.app.pop_screen()


        try:
            self.template=template
            self.textarea.text = template
        except Exception as e:
            print("Failed to update TextArea:", e)
    
    def get_solution_code(self):
        """Return the current code from the editor"""
        return self.query_one(TextArea).text
    
    async def action_run_code(self) -> None:
        """Execute the current code and show results"""
        code = self.query_one(TextArea).text
        all_results=[]
        formatted_results=[]
        match self.language:
            case 'py':
                results=await run_python_code(code, self.challenge)
                self.notify(
                    title="Hey... I started running your code!",
                    message=f"{DAEMON_USER} Wait a sec as I finish!",
                    severity="information",
                    timeout=3,
                    markup=True
                    )
                self.notify(
                    title="Hey mortal...I finished running your code!",
                    message=f"{DAEMON_USER} Check your 'All Tests' tab! See ya~",
                    severity="information",
                    timeout=3,
                    markup=True
                )
                formatted_results = [format_result(result) for result in results]
                self.all_view.update_content(self.challenge, formatted_results)

            case 'js':
                all_results = await run_js_code(code, self.challenge)
                self.notify(
                    title="Hey... I started running your code!",
                    message=f"{DAEMON_USER} Wait a sec as I finish!",
                    severity="information",
                    timeout=3,
                    markup=True
                    )
                formatted_results = [format_result(result) for result in all_results]
                self.notify(
                    title="Hey mortal...I finished running your code!",
                    message=f"{DAEMON_USER} Check your 'All Tests' tab! See ya~",
                    severity="information",
                    timeout=3,
                    markup=True
                )
                self.all_view.update_content(self.challenge, formatted_results)
            case 'cpp':
                self.app.push_screen(self.CompilationStandardPopup(self.challenge, 
                        self.textarea.text, 
                        self.challenge['function_name'], 
                        [test for test in self.challenge['tests'] if not test.get("hidden", False)], 
                        self.language, 
                        self.textarea, 
                        self.all_view))
            case 'java':
                self.app.push_screen(self.CompilationStandardPopup(self.challenge, 
                    self.textarea.text, 
                    self.challenge['function_name'], 
                    [test for test in self.challenge['tests'] if not test.get("hidden", False)], 
                    self.language, 
                    self.textarea, 
                    self.all_view))

    async def action_submit_solution(self):
        """Submit solution for evaluation against test cases, hidden and non-hidden"""
        code = self.query_one(TextArea).text
        namespace={}
        all_results=[]
        formatted_results=[]
        match self.language:
            case 'py':
                results=await run_python_code(code, self.challenge, is_submission=True)
                self.notify(
                    title="Hey... I started running your code!",
                    message=f"{DAEMON_USER} Wait a sec as I finish!",
                    severity="information",
                    timeout=3,
                    markup=True
                    )
                self.all_view.update_submit_content(self.challenge, results)
                self.notify(
                    title="Hey... I finished running your code!",
                    message=f"{DAEMON_USER} Check ur submit tab!",
                    severity="information",
                    timeout=3,
                    markup=True
                    )
            case 'js':
                all_results = await run_js_code(code, self.challenge, True)
                self.notify(
                    title="Hey mortal...I finished running your code!",
                    message=f"{DAEMON_USER} Check ur submit tab!",
                    severity="information",
                    timeout=3,
                    markup=True
                )
                self.all_view.update_submit_content(self.challenge, all_results)
            case 'cpp':
                self.app.push_screen(self.CompilationStandardPopup(self.challenge, 
                    self.textarea.text, 
                    self.challenge['function_name'], 
                    self.challenge['tests'],
                    self.language, 
                    self.textarea, 
                    self.all_view, 
                    is_submission=True))
            case 'java':
                self.app.push_screen(self.CompilationStandardPopup(
                    self.challenge, 
                    self.textarea.text, 
                    self.challenge['function_name'], 
                    self.challenge['tests'],
                    self.language, 
                    self.textarea, 
                    self.all_view, 
                    is_submission=True))

    
    def action_reset_editor(self):
        """Reset the editor to initial state or template"""
        self.app.push_screen(self.EditorResetConfirm(self))

    class CompilationStandardPopup(ModalScreen):
        def __init__(self, chall, user_code, func_name, test_cases, language, editor, testresultswidget, is_submission=False):
            super().__init__()
            self.chall = chall
            self.code = user_code
            self.func_name = func_name
            self.tests=test_cases
            self.all_view = testresultswidget
            self.lang = language
            self.editor = editor
            self.editor_class = Editor
            self.is_submission = is_submission
        async def on_mount(self) -> None:
            if self.lang == "cpp":
                result = None
                try:
                    result = subprocess.run(["clang++", "--version"], capture_output=True, text=True)
                    if result.returncode != 0:
                        try:
                            result = subprocess.run(["g++", "--version"], capture_output=True, text=True)
                        except FileNotFoundError:
                            self.app.pop_screen()
                            self.notify(
                                title="No compiler?",
                                message=f"{DAEMON_USER} Hey, you don't have a supported compiler. How do you want me to compile? Maybe install clang++ or g++?",
                                severity="error",
                                timeout=3,
                                markup=True
                            )
                except FileNotFoundError:
                    try:
                        result = subprocess.run(["g++", "--version"], capture_output=True, text=True)
                    except FileNotFoundError:
                        self.app.pop_screen()
                        self.notify(
                            title="No compiler?",
                            message=f"{DAEMON_USER} Hey, you don't have a supported compiler. How do you want me to compile? Maybe install clang++ or g++?",
                            severity="error",
                            timeout=3,
                            markup=True
                        )
                    except Exception as e:
                        self.app.pop_screen()
                        self.notify(
                            title="Random error?",
                            message=f"{DAEMON_USER} What the heck? Attempted to check g++ version, got {str(e)}",
                            severity="error",
                            timeout=10,
                            markup=True
                        )
                except Exception as e:
                    self.app.pop_screen()
                    self.notify(
                            title="Random error?",
                            message=f"{DAEMON_USER} What the heck? Attempted to check clang++ version, got: {str(e)}",
                            severity="error",
                            timeout=10,
                            markup=True
                        )
            elif self.lang == "java":
                self.jdk_mapping = await asyncio.to_thread(self.scan_jdks) # Store discovered JDKs
                select = self.query_one("#std_select", Select)
                yes_button = self.query_one("#yes_comp", Button)
                no_button = self.query_one("#no_comp", Button)

                options = [(f"Java {v}", v) for v in sorted(self.jdk_mapping.keys(), reverse=True)]
                options.append(("Custom path", "custom")) # Add custom path option

                select.set_options(options)

                if self.jdk_mapping: # If JDKs were found, select the latest
                    select.value = sorted(self.jdk_mapping.keys(), reverse=True)[0]
                else: # Otherwise, default to custom path
                    select.value = "custom"
                
                yes_button.disabled = False
                no_button.disabled = False


        def compose(self) -> ComposeResult:
            with Vertical(id="compilation_dialog"):
                yield Label(f"{DAEMON_USER} Choose ur standard!", id="choose_comp_text")
                if self.lang == "cpp":
                    yield Select([
                    ("C++20", "c++20"),
                    ("C++17", "c++17"),
                    ("C++14", "c++14"),
                    ("C++11", "c++11"),
                    ],
                    value="c++17",
                    id="std_select")
                    with Horizontal(id = "comp_type_select"):
                        yield Button.success("Select", id="yes_comp")
                        yield Button.error("Quit", id="no_comp")
                elif self.lang == "java":
                    yield Select([("Custom", "custom")], value="custom", id="std_select")
                    with Horizontal(id = "comp_type_select"):
                        yield Button.success("Select", id="yes_comp", disabled=True)
                        yield Button.error("Quit", id="no_comp", disabled=True)
        @on(Button.Pressed, "#yes_comp")
        async def confirm_comp(self):
            std_selection=self.query_one("#std_select", Select)
            value=std_selection.value
            if value == "custom":
                self.app.push_screen(Editor.CustomCompilationPath(
                    language=self.lang,
                    editor=self.editor,
                    func_name=self.func_name,
                    tests=self.tests,
                    is_submission=self.is_submission,
                    all_view=self.all_view,
                    chall=self.chall))
                return 
            else:
                if value and isinstance(value, str):
                    if self.lang == "cpp":
                        results = await run_cpp_code(self.code, self.func_name, self.tests, value)
                        self.notify(
                            title="Hey... I started running your code!",
                            message=f"{DAEMON_USER} Wait a sec as I finish!",
                            severity="information",
                            timeout=3,
                            markup=True
                        )
                        formatted_results = [format_result(result) for result in results]
                        if self.is_submission:
                            self.all_view.update_submit_content(self.chall, results)
                            self.notify(
                                title="Hey... I finished running your code!",
                                message=f"{DAEMON_USER} Check ur submit tab!",
                                severity="information",
                                timeout=3,
                                markup=True
                            )
                        else:
                            self.all_view.update_content(self.chall, formatted_results)
                            self.notify(
                                title="Hey mortal...I finished running your code!",
                                message=f"{DAEMON_USER} Check your 'All Tests' tab! See ya~",
                                severity="information",
                                timeout=3,
                                markup=True
                            )

                    elif self.lang == "java":
                        # Ensure 'value' is a key for a discovered JDK path
                        jdk_path = self.jdk_mapping.get(value)
                        if not jdk_path:
                            self.notify(title="Error", message=f"Selected JDK version '{value}' not found in mapping.", severity="error")
                            return

                        results = await run_java_code(self.code, self.func_name, self.tests, jdk_path, self.is_submission)
                        self.notify(
                            title="Hey... I started running your code!",
                            message=f"{DAEMON_USER} Wait a sec as I finish!",
                            severity="information",
                            timeout=3,
                            markup=True
                        )
                        formatted_results = [format_result(result) for result in results]
                        if self.is_submission:
                            self.all_view.update_submit_content(self.chall, results)
                            self.notify(
                                title="Hey... I finished running your code!",
                                message=f"{DAEMON_USER} Check ur submit tab!",
                                severity="information",
                                timeout=3,
                                markup=True
                            )
                        else:
                            self.all_view.update_content(self.chall, formatted_results)
                            self.notify(
                                title="Hey mortal...I finished running your code!",
                                message=f"{DAEMON_USER} Check your 'All Tests' tab! See ya~",
                                severity="information",
                                timeout=3,
                                markup=True
                            )
                    self.app.pop_screen()
                else:
                    self.notify(
                        title="Really?",
                        message=f"{DAEMON_USER} Choose an option! How do you want me to compile if you won't tell me how?",
                        severity="error",
                        timeout=3,
                        markup=True
                    )

        def scan_jdks(self) -> dict:
            system = platform.system()
            jdk_mapping={} 
            if system=="Windows":
                base_dirs = [r"C:\Program Files\Java", r"C:\Program Files (x86)\Java"]
                for base in base_dirs:
                    if os.path.exists(base):
                        for sub in os.listdir(base):
                            jdk_path = os.path.join(base, sub)
                            javac = os.path.join(jdk_path, "bin", "javac.exe")
                            if os.path.exists(javac):
                                try:
                                    result = subprocess.run([javac, "-version"], capture_output=True, text=True)
                                    version_line = result.stdout or result.stderr
                                    if version_line.startswith("javac"):
                                        version_str = version_line.split()[1]
                                        if version_str.startswith("1."):
                                            version = version_str.split(".")[1]
                                        else:
                                            version = version_str.split(".")[0]
                                        jdk_mapping[version] = jdk_path
                                except Exception:
                                    pass
            elif system=="Darwin":
                base_dirs = [r"/Library/Java/JavaVirtualMachines/"]
                for base in base_dirs:
                    if os.path.exists(base):
                        for sub_folder in os.listdir(base):
                            jdk_path = os.path.join(base, sub_folder, "Contents", "Home")
                            javac = os.path.join(jdk_path, "bin", "javac")
                            if os.path.exists(javac):
                                try:
                                    result = subprocess.run([javac, "-version"], capture_output=True, text=True)
                                    version_line = result.stdout or result.stderr
                                    if version_line.startswith("javac"):
                                        version_str = version_line.split()[1]
                                        if version_str.startswith("1."):
                                            version = version_str.split(".")[1]
                                        else:
                                            version = version_str.split(".")[0]
                                        jdk_mapping[version] = jdk_path
                                except Exception:
                                    pass
            elif system=="Linux":
                base_dirs = [r"/usr/lib/jvm"]
                for base in base_dirs:
                    if os.path.exists(base):
                        for sub_folder in os.listdir(base):
                            jdk_path = os.path.join(base, sub_folder)
                            javac = os.path.join(jdk_path, "bin", "javac")
                            if os.path.exists(javac):
                                try:
                                    result = subprocess.run([javac, "-version"], capture_output=True, text=True)
                                    version_line = result.stdout or result.stderr
                                    if version_line.startswith("javac"):
                                        version_str = version_line.split()[1]
                                        if version_str.startswith("1."):
                                            version = version_str.split(".")[1]  
                                        else:
                                            version = version_str.split(".")[0] 
                                        jdk_mapping[version] = jdk_path
                                except Exception:
                                    pass
            self.jdk_mapping = jdk_mapping
            return jdk_mapping


        @on(Button.Pressed, "#no_comp")
        def stop_comp(self):
            self.app.pop_screen()
    class CustomCompilationPath(ModalScreen):       
        def __init__(self, language, editor, func_name, tests, is_submission, all_view, chall):
                    super().__init__()
                    self.language=language
                    self.editor=editor
                    self.func_name = func_name
                    self.tests = tests
                    self.is_submission = is_submission
                    self.all_view = all_view
                    self.chall = chall
                    self.all_view = all_view
                    self.editor_class = Editor

            
        def compose(self) -> ComposeResult:
            with Vertical(id="custom_comp_prompt"):
                placeholder_text = "Enter custom path"
                system = platform.system()
                if self.language == "java":
                    yield Label("Insert custom path (Should be the root of your JDK!)", id="reset_text")
                    if system == "Darwin":
                        placeholder_text = "/Library/Java/JavaVirtualMachines/jdk-21.jdk/Contents/Home"
                    elif system == "Linux":
                        placeholder_text = "/usr/lib/jvm/java-21-openjdk-amd64"
                    elif system == "Windows":
                        placeholder_text = r"C:\Program Files\Java\jdk-21"
                    else:
                        placeholder_text = "Enter JDK root path"
                else:
                    yield Label("Insert custom path", id="reset_text")
                yield Input(placeholder=placeholder_text, id="custom_path_input")
                with Horizontal(id="confirm_custom_buttons"):
                    yield Button.success("Yes", id="yes_custom_button")
                    yield Button.error("No", id="no_custom_button")
    
        async def on_button_pressed(self, event: Button.Pressed) -> None:
            match event.button.id:
                case "yes_custom_button":
                    print(os.path.exists(self.query_one("#custom_path_input", Input).value))
                    input_widget = self.query_one("#custom_path_input", Input)

                    current_path_value = input_widget.value
                    if not current_path_value.strip():
                        self.notify(
                            title="Input needed!",
                            message=f"{DAEMON_USER} [b]Please enter a path![/b]",
                            severity="warning",
                            timeout=5,
                            markup=True
                        )
                        return
                    if not os.path.exists(self.query_one("#custom_path_input", Input).value):
                        self.notify(
                        title="Are you serious?",
                        message=f"{DAEMON_USER} [b]Hey, that path doesn't exist![/b]",
                        severity="error",
                        timeout=5,
                        markup=True
                    ) 
                        return
                    self.app.pop_screen()
                    self.app.pop_screen() #We do it twice since we know there are two layers of screens
                    results = await run_java_code(self.editor.text, self.func_name, self.tests, current_path_value, self.is_submission)
                    self.notify(
                        title="Hey... I started running your code!",
                        message=f"{DAEMON_USER} Wait a sec as I finish!",
                        severity="information",
                        timeout=3,
                        markup=True
                    )
                    formatted_results = [format_result(result) for result in results]
                    if self.is_submission:
                        self.all_view.update_submit_content(self.chall, results)
                        self.notify(
                            title="Hey... I finished running your code!",
                            message=f"{DAEMON_USER} Check ur submit tab!",
                            severity="information",
                            timeout=3,
                            markup=True
                        )
                    else:
                        self.all_view.update_content(self.chall, formatted_results)
                        self.notify(
                            title="Hey mortal...I finished running your code!",
                            message=f"{DAEMON_USER} Check your 'All Tests' tab! See ya~",
                            severity="information",
                            timeout=3,
                            markup=True
                        )
                case "no_custom_button":
                    self.app.pop_screen()
    class EditorResetConfirm(ModalScreen):
        def __init__(self, editor):
            super().__init__()
            self.editor = editor
            
        def compose(self) -> ComposeResult:
            with Vertical(id="reset_confirm_dialog"):
                yield Label("[bold][red]Are you sure you want to reset your code to the template?[/][/]", id="reset_text")
                yield Label("This will also reset any current test results!", id = "test_result_notice")
                with Horizontal(id="reset_buttons"):
                    yield Button.success("Yes", id="yes_reset_button")
                    yield Button.error("No", id="no_reset_button")
    
        def on_button_pressed(self, event: Button.Pressed) -> None:
            match event.button.id:
                case "yes_reset_button":
                    self.editor.textarea.text = self.editor.template
                    self.editor.textarea.refresh()
                    self.editor.all_view.reset_content()
                    self.app.pop_screen()
                case "no_reset_button":
                    self.app.pop_screen()


import argparse
from models import User
from database import SessionLocal
import subprocess

def update_admin(email):
    db = SessionLocal()
    user = db.query(User).filter(User.email == email).first()
    if user:
        user.is_admin = 1 # type:ignore
        db.commit()
        print("User promoted to admin.")
        db.close()
    else:
        print("User not found.")

def main():
    parser = argparse.ArgumentParser(description="Tools used for administrating the Nyxbox backend.")
    parser.add_argument("--promote", action="store_true", help="Promote a user to admin status")
    parser.add_argument("--remove", action="store_true", help="Remove database (DANGER!!)")
    parser.add_argument("email", nargs="?", help="User email address")
    args = parser.parse_args()
    if args.promote:
        if not args.email:
            print("Error: Email is required when using --promote")
            return
        update_admin(args.email)
    elif args.remove:
        # Add your remove database logic here
        result = subprocess.run("rm -rf database.db", shell=True, capture_output=True, text=True)
        if result.returncode == 0:
            print("Command executed successfully")
        else:
            print(f"Command failed: {result.stderr}")
    else:
        parser.print_help()



# Built in libs
from __future__ import annotations
import datetime
import json  # noqa: F401
import os  # noqa: F401
import pathlib
import sys  # noqa: F401
import random

# Third party libs
from sqlalchemy import delete # noqa: F401
from textual import work
from config import Settings  # noqa: F401
from database import SessionLocal, engine  # noqa: F401
from models import Challenges
from sqlalchemy import or_
from sqlalchemy.orm import Session  # noqa: F401
from sqlalchemy.sql.sqltypes import JSON, DateTime, Integer, String
from textual import on  # noqa: F401
from textual.app import App, ComposeResult
from textual.containers import (  # noqa: F401
    Container,
    Horizontal,
    ScrollableContainer,
    Vertical,
)
from textual.reactive import reactive  # noqa: F401
from textual.screen import ModalScreen, Screen  # noqa: F401
from textual.widget import Widget
from textual.widgets import (  # noqa: F401
    Button,
    DataTable,
    Footer,
    Header,
    Input,
    Label,
    ListItem,
    ListView,
    Rule,
    Static,
    TextArea,
)

# Misc stuff
DAEMON_USER="[#B3507D][bold]nyx[/bold][/#B3507D]@[#A3C9F9]hackclub[/#A3C9F9]:~$"
class LabelItem(ListItem):
    def __init__(self, label: str, *, id: str = "") -> None:
        super().__init__(id=id)
        self.label = label
    def compose( self ) -> ComposeResult:
        yield Label(self.label)

class UpdatedDict(): # Call this class, pass it into DictEditScreen, have it modify
    def __init__(self, dict_or_list: "dict | list"):
        self.new_data = dict_or_list
        
class UserChallView(Static):
    def on_mount(self):
        self.styles.align_horizontal = "center"
        NO_CHALL_TEXT=[
            f"{DAEMON_USER} Hm, what an interesting app. Why don't you try pressing something? Maybe 'v'?",
            f"{DAEMON_USER} This better be good since you woke me up. Maybe vend a challenge and see what happens?",
            f"{DAEMON_USER} Welcome Hack Clubbers! (and welcome to everyone else ig). Try vending a challenge!",
            f"{DAEMON_USER} I'm bored...vend a challenge please?"]
        self.update(random.choice(NO_CHALL_TEXT))
    def update_chall(self, chall: dict):
        # Format the profile info
        CHALL_TEXT=[
            f"{DAEMON_USER} Here's your challenge. I hope it's not too hard~",
            f"{DAEMON_USER} Another day, another challenge. You've got this!",
            f"{DAEMON_USER} Let's do this and hopefully learn something. Also, don't fail!",
            f"{DAEMON_USER} Good luck!"
            ]
        if chall.get('difficulty', "N/A").lower() in ["medium", "hard"]:
            CHALL_TEXT=[
            f"{DAEMON_USER} This one might be tough.",
            f"{DAEMON_USER} It's okay to fail! (unless you're a hack clubber of course (im joking!))",
            f"{DAEMON_USER} Bringing on the heat? Let's do it.",
            f"{DAEMON_USER} Good luck! You'll need it."
            ]
        formatted = (
            random.choice(CHALL_TEXT)+"\n"
            f"Name: {chall.get('name', 'N/A')}\n"
            f"Difficulty: {chall.get('difficulty', 'N/A')}\n"
            f"Description: {chall.get('description', 'N/A')}"
        )
        self.update(formatted)
#Screens where we actually manipulate the db
class ChallengeAddScreen(Screen):
    def __init__(self, main_instance: ChallengeListScreen) -> None:
        super().__init__()
        self.main_instance = main_instance
    
    def on_mount(self) -> None:
        pass
    
    def compose(self) -> ComposeResult:
        yield Header()
        yield Input(placeholder="Path to challenge json or folder", id="path_input")
        with Horizontal():
            yield Button.error("Cancel", id="cancel_button")
            yield Button.success("Enter in DB", id="enter_button")
        yield Footer()

    def on_button_pressed(self, event: Button.Pressed):
        match event.button.id: 
            case 'cancel_button':
                self.app.pop_screen()
            case 'enter_button':
                path_input = self.query_one("#path_input", Input)
                chall_path = pathlib.Path(path_input.value)
                db = SessionLocal()
                if not chall_path.exists():
                    self.app.notify(
                        message="Path does not exist.",
                        severity="error"
                    )
                    return
                else:
                    if os.path.isdir(chall_path):
                        for chall in os.listdir(chall_path):
                            with open(chall_path / chall, "r") as file:
                                chall_data = None
                                try:
                                    chall_data = json.load(file)
                                    db = SessionLocal()
                                    chall_exists = db.query(Challenges).filter(
                                        or_(
                                            Challenges.name == chall_data["name"],
                                            Challenges.description == chall_data["description"]
                                            )).first()
                                    if chall_exists:
                                        self.app.notify(
                                            message=f"{chall_data["name"]} already exists, not inserting",
                                            severity="error"
                                        )
                                        continue
                                except json.JSONDecodeError:
                                    self.app.notify(
                                        message="Invalid JSON",
                                        severity="error"
                                    )
                                    return
                            if not chall_data:
                                self.app.notify(
                                        message="Invalid JSON",
                                        severity="error"
                                    )
                                return
                            attribute_names_from_model = Challenges.__mapper__.columns.keys()
                            filtered_data = {k: v for k, v in chall_data.items() if k in attribute_names_from_model}
                            new_chall = Challenges(**filtered_data)
                            db.add(new_chall)
                            db.commit()
                    else:
                        with open(chall_path, "r") as file:
                            chall_data = None
                            try:
                                chall_data = json.load(file)
                            except json.JSONDecodeError:
                                self.app.notify(
                                    message="Invalid JSON",
                                    severity="error"
                                )
                                return
                        if not chall_data:
                            self.app.notify(
                                    message="Invalid JSON",
                                    severity="error"
                                )
                            return
                        # db = SessionLocal()
                        attribute_names_from_model = Challenges.__mapper__.columns.keys()
                        filtered_data = {k: v for k, v in chall_data.items() if k in attribute_names_from_model}
                        new_chall = Challenges(**filtered_data)
                        db.add(new_chall)
                        db.commit()
                    db.close()
                    self.app.notify(
                        message="Successfully updated DB",
                        severity="information"
                    )
                    self.main_instance.load_challenges() # Gotta reload the thing so that we can see the challenges

class DictEditScreen(Screen):
    def __init__(self, edit_json_target: "list | dict", shared_class: UpdatedDict):
        super().__init__()
        self.json_target = edit_json_target
        self.shared_class = shared_class
        # Do we have to unpack it later? 
        # If yes, we set self.single to True (since we have to take the dict out of the list)
        if isinstance(edit_json_target, dict):
            self.single = True
            self.json_target = [edit_json_target]
        elif isinstance(edit_json_target, list) and not all(isinstance(d, dict) for d in edit_json_target):
            self.dismiss(None)
            self.notify("Not JSON, is a list! Manually edit please!")
            self.app.pop_screen()
        else:
            self.single = False
            self.json_target = edit_json_target
        

    def compose(self) -> ComposeResult:
        if isinstance(self.json_target, list):
            all_keys = sorted({k for d in self.json_target for k in d})
            table = DataTable(id="dicts_table")
            table.cursor_type="cell"
            table.add_columns(*all_keys)
            for entry in self.json_target:
                row = [str(entry.get(k, "")) for k in all_keys]
                table.add_row(*row)
            yield table
            yield Input(placeholder="Empty cell", id="edit_cell")
            with Horizontal():
                yield Button("Save Changes", id="save_dict_edits")
                yield Button("Discard Changes", id="discard_dict_edits")
        else:
            self.dismiss(None)
            self.app.notify(message="Invalid type!", severity="error")

    def on_data_table_cell_highlighted(self, event: DataTable.CellHighlighted): 
        # legit don't even know what class event is supposed to be so lets hope and pray textual did it right
        update_input = self.query_one("#edit_cell", Input)
        update_input.value = str(event.value)
        self.current_row = event.cell_key

    def on_input_submitted(self, event: Input.Submitted):
        data_table = self.query_one("#dicts_table", DataTable)
        data_table.update_cell(*self.current_row, event.value, update_width = True)
        self.refresh()

    def on_button_pressed(self, event: Button.Pressed):
        match event.button.id:
            case "save_dict_edits":
                table=self.query_one("#dicts_table", DataTable)
                updated_list = []
                for row in table.ordered_rows:
                    updated_dict = {}
                    for column in table.ordered_columns:
                        cell_val = table.get_cell(row.key, column.key)
                        if cell_val == "" or cell_val is None:
                            continue
                        updated_dict[str(column.label)] = cell_val
                    updated_list.append(updated_dict)
                self.shared_class.new_data = updated_list
                result = updated_list[0] if self.single else updated_list
                self.shared_class.new_data = result
                self.dismiss(result)
            case "discard_dict_edits":
                self.dismiss(None)

class ChallengeEditScreen(Screen):
    def __init__(self, chall_id, main_instance: ChallengeListScreen) -> None:
        super().__init__()
        self.type_mappings = {
            Integer: int,
            String: str,
            JSON: dict,
            DateTime: datetime.datetime
        }
        self.restricted_vals = [
            "id",
            "created_at",
            # "likes",
            # "solves",
            # Temporarily commenting these bcs they'll be important for testing
            "flagged",
            "submitted_by",
            "updated_at"
        ]
        self.chall_id = chall_id
        self.attribute_names_from_model = Challenges.__mapper__.columns.keys()
        self.types_from_model = {
            name: column.type.__class__
            for name, column in Challenges.__mapper__.columns.items()}        
        self.db = SessionLocal()
        self.chall=self.db.query(Challenges).filter(Challenges.id == self.chall_id).first()
        for param in self.attribute_names_from_model:
            if str(param) in self.restricted_vals:
                continue
            else:
                self.last_highlighted_param = str(param)
                break
        self.main_instance = main_instance


    def compose(self) -> ComposeResult:
        self.challenge = self.db.query(Challenges).filter(Challenges.id == self.chall_id).first()
        self.list_view_list = [LabelItem(str(attribute), id=str(attribute)) for attribute in self.attribute_names_from_model if attribute not in self.restricted_vals]
        with Horizontal():
            yield ListView(
                *self.list_view_list,
                 id="params_list" # Yields a LabelItem (see above) for each parameter
            )
            with Vertical():
                val = getattr(self.chall, self.last_highlighted_param or "")
                if val is None:
                    val = ""
                if isinstance(val, (dict, list)):
                    val = json.dumps(val, indent=2)
                yield Label(f"Expected type: {self.types_from_model[self.last_highlighted_param]}", id="edit_expected")
                yield Input(value=val, placeholder="Empty field, input something...", id="input_edit") #This is just a placeholder that gets updated on change
                self.json_edit_btn = Button("Edit in JSON editor", id="json_edit")
                self.json_edit_btn.display = True
                yield self.json_edit_btn
        # with ScrollableContainer():
        #     for attribute in self.attribute_names_from_model:
        #         if str(attribute) in self.restricted_vals:
        #             continue
        #         value = getattr(self.challenge, attribute, "")
        #         yield Input(value=str(value), id=str(attribute))
        yield Button.success("Update attributes", id="update_attrs")
        yield Button.error("Discard changes", id="discard_edit")

    @work
    async def open_json_editor(self):
        try:
            json_val = json.loads(self.query_one("#input_edit", Input).value)
        except json.JSONDecodeError:
            self.app.notify("Not a JSON field!") # Lits just not going to bother
            return
        screen = DictEditScreen(json_val, UpdatedDict(json_val))
        result = await self.app.push_screen_wait(screen)
        if result is not None:
            self.query_one("#input_edit", Input).value = json.dumps(result)
            setattr(self.chall, self.last_highlighted_param, result) # type: ignore

    async def on_button_pressed(self, event:Button.Pressed):
        match event.button.id:
            case "update_attrs":
                if self.chall:
                    self.notify(f"Updated challenge {self.chall.name}")
                    self.app.pop_screen()
                    self.db.commit()
                    self.db.close()
                    self.main_instance.load_challenges()
                    return
                else:
                    self.notify("Chall doesn't seem to exist?")
                    self.app.pop_screen()
                    self.db.rollback()
                    self.db.close()
                    return
            case "discard_edit":
                self.db.rollback()
                self.db.close()
                self.app.pop_screen()
                # for attribute in self.attribute_names_from_model:
                #     attr_value = self.query_one(f"#{str(attribute)}", Input)
            case "json_edit":
                self.open_json_editor()

    def on_list_view_highlighted(self, event:ListView.Highlighted): # We actually have to change this so that it updates the last parameter and then gets the new one instead of updating the current one
        if event:
            # self.json_edit_btn.display = False
            selected_item = event.item # We eventually need to use this to update last_highlighted_param
            label_value = (self.query_one("#input_edit", Input).value)
            label=self.query_one("#input_edit", Input)
            expected_type = self.query_one("#edit_expected", Label)
            expected_type.update(f"Expected type: {self.types_from_model[selected_item.id]}") # type: ignore

            if label_value is None or label_value.strip() == "" or label_value.strip() == "None":
                new_attr_type = None
                setattr(self.chall, self.last_highlighted_param or "", None)
                self.last_highlighted_param = selected_item.id # type: ignore
                label.value = ""
                next_val = getattr(self.chall, selected_item.id) # type: ignore
                if next_val is None:
                    label.value = ""
                elif isinstance(next_val, (dict, list)):
                    label.value = json.dumps(next_val, indent=2)
                    # self.json_edit_btn.display = True
                    self.json_edit_btn.refresh()
                else:
                    label.value = str(next_val)
                    # self.json_edit_btn.display = False
                return
            if hasattr(self.chall, self.last_highlighted_param): # type: ignore
                #self.chall.last_highlighted_param = label_value # Need to use setattr here bcs its trying to literally set a param called last_highlighted_param
                try:
                    attr_type = self.types_from_model.get(self.last_highlighted_param)
                    new_attr_type = self.type_mappings.get(attr_type, None)
                    if new_attr_type:
                        if new_attr_type is dict:
                            # self.json_edit_btn.display=True
                            if isinstance(label_value, str):
                                try:
                                    label_value = json.loads(label_value)
                                except json.JSONDecodeError:
                                    list_view = self.query_one("#params_list", ListView)
                                    list_view.index  = next(
                                        i for i, item in enumerate(self.list_view_list)
                                        if item.id == self.last_highlighted_param
                                    )
                                    self.app.notify("Invalid JSON!", severity="error")
                                    return
                        else:
                            pass
                            # self.json_edit_btn.display=False
                        if (isinstance(label_value, list) and all(isinstance(d, dict) for d in label_value)) or isinstance(label_value, list):
                            new_attr_type = list
                            # self.json_edit_btn.display=True
                        elif isinstance(label_value, dict):
                            new_attr_type = dict
                            # self.json_edit_btn.display=True

                        # else:
                        #     self.app.notify(
                        #         message="The type is wrong?", 
                        #         severity="error")
                        elif isinstance(new_attr_type, datetime.datetime):
                            try:
                                label_value=datetime.datetime.fromisoformat(label_value)
                            except ValueError:
                                list_view = self.query_one("#params_list", ListView)
                                list_view.index  = next(
                                    i for i, item in enumerate(self.list_view_list)
                                    if item.id == self.last_highlighted_param
                                )
                                self.app.notify("Invalid time!", severity="error")
                                return
                        else:
                            label_value = new_attr_type(label_value) # type: ignore
                except (ValueError, KeyError, TypeError):
                    list_view = self.query_one("#params_list", ListView)
                    list_view.index  = next(
                        i for i, item in enumerate(self.list_view_list)
                        if item.id == self.last_highlighted_param
                    )
                    self.app.notify("Attribute doesn't seem to exist or conversion failed", severity="error")
                    # self.last_highlighted_param = selected_item.label
                    return
                setattr(self.chall, self.last_highlighted_param, label_value) # type: ignore
                self.last_highlighted_param = selected_item.id # type: ignore
                val = getattr(self.chall, selected_item.id) # type: ignore
                if isinstance(val, (dict, list)):
                    label.value = json.dumps(val)
                else:
                    label.value = str(val)
            else:
                self.app.notify("Attribute doesn't seem to exist?")
                self.last_highlighted_param = selected_item.id # type: ignore
                return

class ApprovalScreen(Screen): 
    # This screen is going to be mostly identical to ChallengeListScreen
    # Except we're gonna steal something from the main app >:3 (aka the frontend/tui app/client/idek)
    # This is more admin specific so that we can approve user submitted challenges

    def __init__(self, main_instance: ChallengeListScreen) -> None:
        super().__init__()
        self.main_instance = main_instance

    def on_mount(self) -> None:
        # display_table = self.query_one("#challenge_table", DataTable)
        # db = SessionLocal()
        self.load_challenges()
        self.second_pass_approve = None
        self.second_pass_deny = None
        self.attempted_row = None
        # self.main_instance

    def compose(self) -> ComposeResult:
        yield Header()
        with Vertical():
            yield DataTable(id="approval_table", cursor_type="row")
            with Horizontal():
                yield Button(id="approve_chall_approval", label="Approve challenge")
                yield Button(id="deny_chall_approval", label="Deny challenge")
                yield Button.error(id="exit_approval", label="Exit")
            yield UserChallView(id="user_chall_view_approve")
        yield Footer()

    def load_challenges(self):
        display_table = self.query_one("#approval_table", DataTable)
        display_table.clear()
        with SessionLocal() as db:
            chall_query = db.query(Challenges).filter(
                or_(
                    Challenges.is_approved == 0,
                    Challenges.flagged == 1,))
            chall_all = chall_query.all()
            if len(chall_all) == 0:
                self.app.pop_screen()
                self.app.notify("No flagged or non-approved challenges left!")                
            try:
                attribute_names_from_model = Challenges.__mapper__.columns.keys()
            except AttributeError:
                print("Error: Could not retrieve column names from Challenges model.")
                self.app.notify("Error: Could not retrieve column names from Challenges model.")
                attribute_names_from_model = ["id", "name", "error"] 
            display_headers = [name.replace("_", " ").title() for name in attribute_names_from_model]
            display_table.clear(columns=True) 
            display_table.add_columns(*display_headers)
            for chall in chall_all:
                row = [getattr(chall, attr) for attr in attribute_names_from_model]
                display_table.add_row(*row)

    def unset_param(self, param):
        param = False

    def on_data_table_row_highlighted(self, event: DataTable.RowHighlighted):
        table = self.query_one("#approval_table", DataTable)
        chall_info_display = self.query_one("#user_chall_view_approve", UserChallView)
        # chall = event.row_key
        chall = table.get_row(event.row_key)
        chall_dict = {"name": chall[1], "difficulty": chall[7], "description": chall[2]}
        chall_info_display.update_chall(chall_dict)
        # update_input = self.query_one("#edit_cell", Input)
        # update_input.value = str(event.value)

    def on_button_pressed(self, event: Button.Pressed):
        match event.button.id:
            case "approve_chall_approval":
                table = self.query_one("#approval_table", DataTable)
                row = table.get_row_at(table.cursor_row)
                if self.second_pass_approve and self.attempted_row == row:
                    with SessionLocal() as db:
                        chall = db.query(Challenges).filter(
                            Challenges.id == row[0], Challenges.is_approved != 1, Challenges.name == row[1]
                        ).first()
                        chall.is_active = 1 #type: ignore
                        chall.is_approved = 1 #type: ignore
                        chall.is_reviewed = 1 #type: ignore
                        db.commit()
                        # db.close()
                    self.app.notify(f"Challenge {row[1]}, id {row[0]} has been approved.")
                    self.load_challenges()
                else:
                    self.attempted_row = row
                    self.app.notify(f"Are you sure you want to approve {row[1]}, id {row[0]}? Click again if yes")
                    self.second_pass_approve = True
                    self.set_timer(3, self.unset_param(self.second_pass_approve))
            case "deny_chall_approval":
                table = self.query_one("#approval_table", DataTable)
                row = table.get_row_at(table.cursor_row)
                if self.second_pass_deny and self.attempted_row == row:
                    with SessionLocal() as db:
                        chall = db.query(Challenges).filter(
                            Challenges.id == row[0], Challenges.name == row[1]
                        ).first()
                        chall.is_approved = 0 #type:ignore just in case it wasn't alr 0
                        chall.is_reviewed = 1 #type: ignore
                        chall.is_active = 0 #type: ignore js in case :3
                        db.commit()
                        # db.close()
                    self.app.notify(f"Challenge {row[1]}, id {row[0]} has been denied.")
                    self.load_challenges()
                else:
                    self.attempted_row = row
                    self.app.notify(f"Are you sure you want to deny {row[1]}, id {row[0]}? Click again if yes")
                    self.second_pass_deny = True
                    self.set_timer(3, self.unset_param(self.second_pass_approve))
            case "exit_approval":
                self.app.pop_screen()
                self.main_instance.load_challenges()
                

class ChallengeListScreen(Screen):
    def on_mount(self) -> None:
        # display_table = self.query_one("#challenge_table", DataTable)
        # db = SessionLocal()
        self.load_challenges()
        self.second_pass = False

    def compose(self) -> ComposeResult:
        yield Header()
        with Vertical():
            yield DataTable(id="challenge_table", cursor_type="row")
            with Horizontal():
                yield Button(id="add_chall", label="Add challenge")
                yield Button(id="edit_chall", label="Edit challenge")
                yield Button(id="remove_chall", label="Remove challenge")
                yield Button(id="approve_chall", label="Approve Challenges")
        yield Footer()
    def load_challenges(self):
        display_table = self.query_one("#challenge_table", DataTable)
        display_table.clear()
        with SessionLocal() as db:
            chall_all = db.query(Challenges).all()
            try:
                attribute_names_from_model = Challenges.__mapper__.columns.keys()
            except AttributeError:
                print("Error: Could not retrieve column names from Challenges model.")
                self.app.notify("Error: Could not retrieve column names from Challenges model.")
                attribute_names_from_model = ["id", "name", "error"] 
            display_headers = [name.replace("_", " ").title() for name in attribute_names_from_model]
            display_table.clear(columns=True) 
            display_table.add_columns(*display_headers)
            for chall in chall_all:
                row = [getattr(chall, attr) for attr in attribute_names_from_model]
                display_table.add_row(*row)

    def unset_second_pass(self):
        # This is literally only necessary because textual
        # won't let you run raw python in the set_timer func
        self.second_pass = False

    def on_button_pressed(self, event: Button.Pressed):
        match event.button.id:
            case "add_chall":
                self.app.push_screen(ChallengeAddScreen(self)) # type: ignore
            case "edit_chall":
                datatable = self.query_one("#challenge_table", DataTable)
                current_row = datatable.get_row_at(datatable.cursor_row)
                with SessionLocal() as db:
                    chall = db.query(Challenges).filter(Challenges.id == current_row[0]).first()
                    if chall:
                        chall_id = chall.id
                    else:
                        self.app.notify(
                            message="Uh oh, the chall doesn't seem to exist in the DB?",
                            severity="error"
                        )
                        return
                self.app.push_screen(ChallengeEditScreen(chall_id, self)) 

            case "remove_chall":
                table = self.query_one("#challenge_table", DataTable)
                row = table.get_row_at(table.cursor_row)
                if self.second_pass:
                    with SessionLocal() as db:
                        chall = db.query(Challenges).filter(
                                row[0] == Challenges.id
                        ).first()
                        db.delete(chall)
                        db.commit()
                        self.app.notify(f"Challenge {row[1]} with id {row[0]} successfuly deleted!")
                        self.second_pass = False
                        self.load_challenges()
                    # delete_query = delete(Challenges).where(row[0] == Challenges.id)
                else:
                    self.notify(f"Are you sure you want to delete {row[1]} with id {row[0]}? Click again if yes.", timeout=3)
                    self.second_pass = True
                    self.set_timer(10, self.unset_second_pass)
            case "approve_chall":
                self.app.push_screen(ApprovalScreen(self))
    
class DBManagement(App):
    def on_mount(self):
        self.push_screen(ChallengeListScreen())
        
def main():
    app = DBManagement()
    app.run()

if __name__ == "__main__":
    main()


import json
import random
from importlib.resources import files
from pathlib import Path

#TODO: Actually make caching because this just isn't going to work out
def vend_random_chall(challenge_list):
    # challenge_files = list(challenge_dir.iterdir())
    # chosen = random.choice(challenge_files)
    # with chosen.open("r", encoding="utf-8") as f:
    #     return json.load(f)
    return random.choice(challenge_list)
    
def list_all_chall():
    challenge_dir = files("nyxbox").joinpath("../challenges")
    return [Path(f.name).stem for f in challenge_dir.iterdir() if f.name.endswith(".json")]

def list_all_chall_detailed():
    challenge_dir = files("nyxbox").joinpath("../challenges")
    return [
        json.load(f.open("r", encoding="utf-8"))
        for f in challenge_dir.iterdir()
        if f.name.endswith(".json")
    ]

def return_tests(challenge):
    challenge_dir = files("nyxbox").joinpath("../challenges")
    path = challenge_dir.joinpath(f"{challenge}.json")
    with path.open("r", encoding="utf-8") as f:
        return json.load(f)["tests"]

